/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.view.generic;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.jface.viewers.ViewerComparator;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.ui.annotation.AnnotationEntry;
import org.fortiss.tooling.kernel.model.INamedElement;

/**
 * Comparator that sorts rows according to hierarchy level. Elements within one hierarchy level
 * are sorted by name.
 */
public class HierarchicalNameComparator extends ViewerComparator {

	/** {@inheritDoc} */
	@Override
	public int compare(Viewer viewer, Object o1, Object o2) {
		if(o1 instanceof AnnotationEntry && o2 instanceof AnnotationEntry) {
			IModelElement modelElement1 = ((AnnotationEntry)o1).getModelElement();
			IModelElement modelElement2 = ((AnnotationEntry)o2).getModelElement();
			int modelElement1Level = GenericAnnotationView.getModelElementLevel(modelElement1);
			int modelElement2Level = GenericAnnotationView.getModelElementLevel(modelElement2);

			EObject e1 = modelElement1;
			EObject e2 = modelElement2;

			if(modelElement1Level < modelElement2Level) {
				// Element1 is higher than element2: Get ancestor of element2 that is at the
				// level as the element1.
				EObject ancestor =
						GenericAnnotationView.getModelElementAncestor(modelElement2,
								modelElement2Level - modelElement1Level);

				if(modelElement1 == ancestor) {
					// Element1 is offspring of element2
					return -1;
				}

				// Further comparison starts on elements at the same level
				e2 = ancestor;
			} else if(modelElement1Level > modelElement2Level) {
				// Element2 is higher than element1: Get ancestor of element1 that is at the
				// level as the element2.
				EObject ancestor =
						GenericAnnotationView.getModelElementAncestor(modelElement1,
								modelElement1Level - modelElement2Level);

				if(modelElement2 == ancestor) {
					// Element2 is offspring of element1
					return 1;
				}

				// Further comparison starts on elements at the same level
				e1 = ancestor;
			}
			// Find direct children of common root element of element1 and element2...
			do {
				if(e1.eContainer() == e2.eContainer()) {
					break;
				}
				e1 = e1.eContainer();
				e2 = e2.eContainer();
			} while(true);

			// ... and compare them by name...
			if(e1 instanceof INamedElement && e2 instanceof INamedElement) {
				// Elements are at the same level -> compare by name
				return(((INamedElement)e1).getName().compareTo(((INamedElement)e2).getName()));
			}
		}

		// ... or hash code (as a fallback ordering)
		// int Integer.compare(int x, int y) requires JDK 1.7 which is not available
		// on the build server
		return new Integer(o1.hashCode()).compareTo(o2.hashCode());
	}
}
